/* **************************************************************************
   ARDUINO DATA LOGGER CODE
   
   Version: Z11_Optimised (final)
   Date: 29.09.2016
   
   Copyright 2016 by Fabian Tobias Mayer (fabianmayer@web.de)
   This code may not be used for purposes which are not related to 
   the master thesis of Fabian Tobias Mayer without his prior
   written consent. 
   **************************************************************************

   Function Summary: 
   - Warm up for the multiplexers
   - Adjustment of the reference pin of the instrumentational amplifiers
   - Setting time for the instrumentational amplifiers (still multiplexing)
   - Set the E and S0…S2 pins HIGH and LOW to govern the multiplexers
   - Read the measurement value for the specific multiplexer settings
   - Store the values on the SD card (return error when it is not possible to    
     initialise or write on the SD card)
   - Determine the gain state (return error when gain state is changed)
   - Start and stop the logging on the SD card when the start and stop button
     is pushed (light the green LED when measuring)
   - Automatically start a new file when there are more than 10 000 samples
     in the file (to keep the opening and closing time low
   - Differently increasing file names when restarting manually or 
     automatically new files
   - Stop the Measurement when there is an error and light the red error LED

   Function *Setup*:
   - Set error and measurement LED high to show that the setup is running
   - Return an error when the SD card is missing and switch the green LED off
   - Deactivate all Multiplexers by setting the E pins HIGH
   - Determine the initial gain state
   - Set error and measurement LED low to show that the setup is done and the
     system is ready to log
   - Determine the highest file name on the SD card not to overwrite existing 
     files
   - Make the multiplexers to warm up   
   - Determine the maximum voltage of the channels and adjust it to 1.9 V 
     (the reference voltage of the instrumentation amplifier is not raised 
     higher than 1.9 V; the adjustment is only done during an allowed 
     adjustment time)
   - There is a setting time for the instrumentational amplifier before the 
     logging is enabled which is signalised by switching off the error and 
     measurement LED

   Function *Loop*:
   - Determine the current gain state and set errorState true when it changed
   - Actions which should be taken if the errorState is true: Light the 
     errorLed, set measuring state false
   - Determine the start stop button state and de-bounce it with 30ms
   - Set the MSState false or true dependant whether a measurement is running
   - When the MSState was set true the fileNo (first three numbers of the 
     File name) is increased to write to a new file when the next measurement 
     is started
   - The sample counter is reset when sampling on a new file is started 
     manually
   - When the maximum number of samples per file is reached a new file is
     Started (The last three digits of the file name are increased)
   - The file sample counter is reset when a new measurement file is
     started automatically
   - Set the E pins that the first multiplexer set is enabled and the other
     two disabled
   - Go through the S pin HIGH/LOW combinations to connect the channels to
     the output after each other (there is a delay between the channels in 
     order to grant stability of the signal before reading it)
   - Adapt and read all inputs of the first multiplexer set
   - Disable the first multiplexer set and do the same procedure for the 
     second and third multiplexer set
   - Write the measured data of all channels on the SD card when MSstate true
   - The data is logged and sampling counter increases as long as the file 
     name is the same

   Connections:
   - Pin 2 (OUTPUT) connected to the inhibit Pin E of multiplexer 1+ and 1-  
     !!!Requires external pull down resistors!!!
   - Pin 3 (OUTPUT) connected to the inhibit Pin E of multiplexer 2+ and 2- 
     !!!Requires external pull down resistors!!!
   - Pin 4 (OUTPUT) connected to the inhibit Pin E of multiplexer 3+ and 3- 
     !!!Requires external pull down resistors!!!
   - Pin 5 (OUTPUT) connected to the S0 Pin of all multiplexes 
     !!!Requires external pull down resistors!!!
   - Pin 6 (OUTPUT) connected to the S1 Pin of all multiplexes
     !!!Requires external pull down resistors!!!
   - Pin 7 (OUTPUT) connected to the S2 Pin of all multiplexes
     !!!Requires external pull down resistors!!!
   - Pin 8 (INPUT) connected to the gain low side of the gain state switch 
     !!!Requires external pull down resistor!!!
     (INPUT_PULLUP cannot be used as this pin is never connected to ground)
   - Pin 9 (INPUT_PULLUP) connected to the button to start and stop the 
     measurement (connected to ground when button pushed)
   - Pin 10 (OUTPUT) connected to the green LED which is lighted, when data 
     is logged
   - Pin 11 (OUTPUT) connected to the red error LED which is lighted when 
     there is an error
   - Pin 12 (-) used for the chip select (CS) of the SPI communication with 
     the SD card
   - Pin 13 (INPUT_PULLUP) connected to the detection switch of the SD card,
     (connected to ground, when the SD card is present)
   - Pin A0 (OUTPUT) connected to the reference terminal of the 
     instrumentational amplifier
   - Pin A1 (INPUT) connected to the output of the second instrumentation 
     Amplifier (more accurate with external pull down than with internal pull 
     up resistor) !!!Requires external pull down resistor!!!
   - SPI (-) connector connected to the SD card
   - Reset Pin (-) connected to a button which connects to ground when pushed

   User Interface:
   - Both LEDs are lighted during the setup. When both LEDs are of the setup 
     and calibration is done and data can be logged
   - Start/stop button: Start and stop the data logging on the SD card
   - Green measurement LED: If the green LED is lighted data is logged
   - Red error LED: If the red LED is lighted the user needs to go through
     the error check list
   - Reset button: To be pressed to restart the system after an error was
     corrected
   - Name of the file the data is logged on is aaabbb.txt 
     (aaa is increasing from 001 to 999 when the user starts and stops a
     measurement; bbb is increasing from 001 to 999 when a new file is
     started after 10 000 samples)

   Error State: (Reset needed to get out of the error state)
   --> Measurement is stopped
   --> Green measurement LED is switched off
   --> Red errorLED is lighted
   - Error when the SD card is missing when the measuring system is powered
   - Error when gain state is changed after the Arduino is started up
   - Error when the SD card cannot be initialised (no SD card available / SD 
     card broken
   - Error when it cannot be written on the SD card any more (full/ removed/ 
     broken)
   - Error when the fileNo (more than 999 measurements started by the user) 
     or measNo (more than 999 sub files started by the Arduino)

*****************************************************************************
*/


//_______________________________Libraries___________________________________
#include <SPI.h>
#include <SD.h>


//_______________________________Definitions_________________________________

//-----Define Pins-----
//Multiplexer Governing:
#define E1Pin 2             //Select the output pin to enable mux 1+ and 1-
#define E2Pin 3             //Select the output pin to enable mux 2+ and 2-
#define E3Pin 4             //Select the output pin to enable mux 3+ and 3-
#define S0Pin 5             //Select the output pin for steering the mux 
                            //output S0
#define S1Pin 6             //Select the output pin for steering the mux 
                            //output S1
#define S2Pin 7             //Select the output pin for steering the mux 
                            //output S2

//Reference adjustment:
#define refAdjPin A0        //Select the output to adjust the 
                            //instrumentational amplifier with at the 
                            //reference terminal

//Determine gain state:
#define GSPin 8             //Select the input pin to determine the gain 
                            //state (connected to +3.3V when common forces)

//Start/stop logging on the SD card:
#define measSSPin 9          //Select the input pin to start and stop the 
                             //measurement (connected to 0V when pushed)
#define ledMSPin 10          //Select the output pin for the LED which shows 
                             //whether data is logged

//Error:
#define errorLedPin 11      //Select the output pin for the errorLED

//Write to SD card:
const int chipSelect = 12;

//Check presence of SD card:
#define sdDetectPin 13      //Select the input pin to determine whether the 
                            //SD card is present 
                            //(connected to 0V when present)

//Determine Measrued Values:
#define MSPin A1            //Select the input pin for the measurement signal


//-----Define Variables-----
//Warm up and reference adjustment:
unsigned long startWU = 0;    //Variable to store the time [ms] when the 
                              //warm up begins
int WUTime = 2000;            //Time [ms] to warm up before starting with the 
                              //adjustment of the instrumentational amplifier
int adjTime = 10000;          //Time [ms] how long adjustment should be done
int chMax = 0;                //Variable to store the maximal voltage
                              //[0...4095] output of the channels 1 to 24
int vSenseMax = 2482;         //Voltage 2V [0...4095] the highest output 
                              //channel of the instrumentational amplifier 
                              //should not be adjusted over (vAdj/3.3*4096)
int vAdjInc = 2;              //Voltage increment [0...1023] with which the 
                              //voltage is increased for the adjustment
int vAdjMax = 590;            //Voltage which is maximal apply on the
                              //reference pin 1.9V (not more than 2V allowed 
                              //by the supplier)
int vAdj = 0;                 //Variable to store the voltage [0...1023] 
                              //applied to the reference pin of the 
                              //instrumentation amplifier
int settingTime = 3000;       //Time [ms] required until the output voltage 
                              //is stable (changes less than 3 uV)

//Naming of the data files:
String fileName = "";         //Variable to store the file number
int fileNo = 1;               //Variable for the file name for a new 
                              //measurement started by the user
int measNo = 1;               //Variable for the file name for a new 
                              //measurement started by the programme

//Automatic new sub file:
int sampleNo = 0;             //Variable to count the samples in one file
int sampleNoMax = 10000;      //Variable how many samples should be written 
                              //to one file

//Error:
boolean errorState = false;   //Variable to store whether there is an error

//Start and stop logging on the SD card:
boolean buttonState = false;  //Variable to store the button state
boolean msState = false;      //Variable to store whether logging to the SD 
                              //card should be done

//Multiplexer steering:
int tSet = 100;               //Setting time in microseconds (10^-6) after 
                              //switching before the measurement is done
                              //Setting time multiplexers: 400-1000ns
                              //Setting time instrumentation amplifiers:15us 
                              //Output voltage drop when higher than 10kHz 
                              //--> 100 micro seconds

//Determine Measured Values:
int ch01 = 0;            //Variable to store the measured data of channel 1
int ch02 = 0;            //Variable to store the measured data of channel 2
int ch03 = 0;            //Variable to store the measured data of channel 3
int ch04 = 0;            //Variable to store the measured data of channel 4
int ch05 = 0;            //Variable to store the measured data of channel 5
int ch06 = 0;            //Variable to store the measured data of channel 6
int ch07 = 0;            //Variable to store the measured data of channel 7
int ch08 = 0;            //Variable to store the measured data of channel 8
int ch09 = 0;            //Variable to store the measured data of channel 9
int ch10 = 0;            //Variable to store the measured data of channel 10
int ch11 = 0;            //Variable to store the measured data of channel 11
int ch12 = 0;            //Variable to store the measured data of channel 12
int ch13 = 0;            //Variable to store the measured data of channel 13
int ch14 = 0;            //Variable to store the measured data of channel 14
int ch15 = 0;            //Variable to store the measured data of channel 15
int ch16 = 0;            //Variable to store the measured data of channel 16
int ch17 = 0;            //Variable to store the measured data of channel 17
int ch18 = 0;            //Variable to store the measured data of channel 18
int ch19 = 0;            //Variable to store the measured data of channel 19
int ch20 = 0;            //Variable to store the measured data of channel 20
int ch21 = 0;            //Variable to store the measured data of channel 21
int ch22 = 0;            //Variable to store the measured data of channel 22
int ch23 = 0;            //Variable to store the measured data of channel 23
int ch24 = 0;            //Variable to store the measured data of channel 24

//Determine gain state:
int startGS = 0;         //Variable of the gain state when stating the 
                         //system (1 for low loads; 2 for high loads)
int nowGS = 0;           //Variable of the current gain state 
                         //(1 for common forces; 2 for maximum forces)


//******************************** SETUP ************************************


void setup() {

  //-----Serial Interface-----

  Serial.begin(115200);


  //-----Pins Settings-----

  //Multiplexer steering:
  pinMode(E1Pin, OUTPUT);      //Define the pin 2 as OUTPUT 
  pinMode(E2Pin, OUTPUT);      //Define the pin 3 as OUTPUT 
  pinMode(E3Pin, OUTPUT);      //Define the pin 4 as OUTPUT 
  pinMode(S0Pin, OUTPUT);      //Define the pin 5 as OUTPUT 
  pinMode(S1Pin, OUTPUT);      //Define the pin 6 as OUTPUT 
  pinMode(S2Pin, OUTPUT);      //Define the pin 7 as OUTPUT 

  //Adjustment of the reference terminal:
  pinMode(refAdjPin, OUTPUT);  //Define pin A0 as output
  analogWriteResolution(10);   //Set the analogWrite resolution to its 
                               //maximum at 10-bit

  //Gain State:
  pinMode(GSPin, INPUT);       //Define the pin 8 as INPUT 
                               //(INPUT_PULLUP here not possible as this pin 
                               //is never connected to ground)

  //Start and stop logging on the SD card:
  pinMode(measSSPin, INPUT_PULLUP);    //Define pin 9 as INPUT_PULLUP
  pinMode(ledMSPin, OUTPUT);           //Define pin 10 as OUTPUT

  //Error:
  pinMode(errorLedPin, OUTPUT);        //Define the pin 11 as OUTPUT

  //SD Card detection:
  pinMode(sdDetectPin, INPUT_PULLUP);  //Define pin 11 as INPUT_PULLUP

  //Measure:
  pinMode(MSPin, INPUT);              //Define the pin A1 as INPUT 
  analogReadResolution(12);           //Set the analogue read resolution of
                                      //the Arduino Zero to the max at 12-bit


  //-----Disable Multiplexers-----

  digitalWrite(E1Pin, HIGH);
  digitalWrite(E2Pin, HIGH);
  digitalWrite(E3Pin, HIGH);


  //-----Signalise Setup Is Running-----

  digitalWrite(ledMSPin, HIGH);
  digitalWrite(errorLedPin, HIGH);


  //-----Error Mode When SD Card Missing-----

  if (digitalRead(sdDetectPin) == HIGH) {
    errorState = true;
    digitalWrite(ledMSPin, LOW);
    Serial.println("SD card missing in Setup");
  }


  //-----Determine the Gain State-----

  //Low loads:
  if (digitalRead(GSPin) == HIGH) {
    startGS = 1;
  }
  //High loads:
  else if (digitalRead(GSPin) == LOW) {
    startGS = 2;
  }


  //-----Initialise SD Card-----
 
  Serial.print("Initializing SD card...");

  //Check if the card is present and can be initialised:
  if (!SD.begin(chipSelect)) {
    errorState = true;
    Serial.println("Card failed, or not present");
  }
  Serial.println("card initialised.");


  //-----Prevent File Overwriting-----

  //Generate the lowest file name possible:
  fileName = "00" + String(fileNo) + "00" + String(measNo) + ".txt";    

  //Increase file number until a file name is generated which does not exist:
  //(This number is kept than to sample on)
  while (SD.exists(fileName)) {
    fileNo = fileNo + 1;
    
    //Generate the file name:
    if (fileNo < 10) {                              //fileNo [0,9]
      fileName = "00" + String(fileNo) + "00" + String(measNo) + ".txt";
    }
    else if (fileNo < 100 && fileNo >= 10) {        //fileNo [10,99]
      fileName = "0" + String(fileNo) + "00" + String(measNo) + ".txt";
    }
    else if (fileNo < 1000 && fileNo >= 100) {      //fileNo [100,999]
      fileName = String(fileNo) + "00" + String(measNo) + ".txt";
    }

    //Send an error if there are already more than 999 manually started 
    //measurements on the SD card
    else {                                          //fileNo >999
      Serial.println("More than 999 files on the card");
      errorState = true;
    }
  }


  //_______________Multiplexer Warm Up and Reference Adjustment______________

  //Store the time when the warm up is started:
  startWU = millis();

  //Conduct the warm up of the multiplexers, adjust the instrumentation 
  //amplifier and let the system a set after the adjustment:
  while (millis() < (startWU + WUTime + adjTime + settingTime)) {

    //-----Multiplexing-----
    Serial.println("Mux warm up and Ref adjust");

    //-----Multiplexer Set 1-----
    //Enable Multiplexer Set 1:
    digitalWrite(E1Pin, LOW);

    //A0 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch01 = analogRead(MSPin);

    //A1 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch02 = analogRead(MSPin);

    //A2 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch03 = analogRead(MSPin);

    //A3 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch04 = analogRead(MSPin);

    //A4 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch05 = analogRead(MSPin);

    //A5 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch06 = analogRead(MSPin);

    //A6 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch07 = analogRead(MSPin);

    //A7 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch08 = analogRead(MSPin);

    //Disable Multiplexer 1
    digitalWrite(E1Pin, HIGH);

    //-----Multiplexer Set 2-----
    //Enable Multiplexer Set 2:
    digitalWrite(E2Pin, LOW);

    //A0 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch09 = analogRead(MSPin);

    //A1 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch10 = analogRead(MSPin);

    //A2 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch11 = analogRead(MSPin);

    //A3 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch12 = analogRead(MSPin);

    //A4 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch13 = analogRead(MSPin);

    //A5 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch14 = analogRead(MSPin);

    //A6 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch15 = analogRead(MSPin);

    //A7 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch16 = analogRead(MSPin);

    //Disable Multiplexer 1
    digitalWrite(E2Pin, HIGH);

    //-----Multiplexer Set 3-----
    //Enable Multiplexer Set 3:
    digitalWrite(E3Pin, LOW);

    //A0 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch17 = analogRead(MSPin);

    //A1 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch18 = analogRead(MSPin);

    //A2 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch19 = analogRead(MSPin);

    //A3 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, LOW);
    delayMicroseconds(tSet);
    ch20 = analogRead(MSPin);

    //A4 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch21 = analogRead(MSPin);

    //A5 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, LOW);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch22 = analogRead(MSPin);

    //A6 Mux 1 connected:
    digitalWrite(S0Pin, LOW);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch23 = analogRead(MSPin);

    //A7 Mux 1 connected:
    digitalWrite(S0Pin, HIGH);
    digitalWrite(S1Pin, HIGH);
    digitalWrite(S2Pin, HIGH);
    delayMicroseconds(tSet);
    ch24 = analogRead(MSPin);

    //Disable Multiplexer 1
    digitalWrite(E3Pin, HIGH);

    //-----Adapt the Reference Pin-----
    //Calculate the current maximum output of the channels 1 to 24 [0…4096]:
    chMax = 0;        //Set chMax to 0 that it can be increased again for the
                      //following loop
    if (chMax < ch01) {
      chMax = ch01;
    }
    if (chMax < ch02) {
      chMax = ch02;
    }
    if (chMax < ch03) {
      chMax = ch03;
    }
    if (chMax < ch04) {
      chMax = ch04;
    }
    if (chMax < ch05) {
      chMax = ch05;
    }
    if (chMax < ch06) {
      chMax = ch06;
    }
    if (chMax < ch07) {
      chMax = ch07;
    }
    if (chMax < ch08) {
      chMax = ch08;
    }
    if (chMax < ch09) {
      chMax = ch09;
    }
    if (chMax < ch10) {
      chMax = ch10;
    }
    if (chMax < ch11) {
      chMax = ch11;
    }
    if (chMax < ch12) {
      chMax = ch12;
    }
    if (chMax < ch13) {
      chMax = ch13;
    }
    if (chMax < ch14) {
      chMax = ch14;
    }
    if (chMax < ch15) {
      chMax = ch15;
    }
    if (chMax < ch16) {
      chMax = ch16;
    }
    if (chMax < ch17) {
      chMax = ch17;
    }
    if (chMax < ch18) {
      chMax = ch18;
    }
    if (chMax < ch19) {
      chMax = ch19;
    }
    if (chMax < ch20) {
      chMax = ch20;
    }
    if (chMax < ch21) {
      chMax = ch21;
    }
    if (chMax < ch22) {
      chMax = ch22;
    }
    if (chMax < ch23) {
      chMax = ch23;
    }
    if (chMax < ch24) {
      chMax = ch24;
    }

    //Increase the voltage:
    //- if the warm up time is over
    //- if the maximal adjustment time is not over
    //- if the voltage on the output is not higher than vSenseMax
    //- if the maximum on the reference pin is lower than the allowed vAdjMax 
    if (millis() > (startWU + WUTime) && millis() < (startWU + WUTime +
        adjTime) && vAdj < (vAdjMax + vAdjInc) && chMax < vSenseMax) {
      vAdj = vAdj + vAdjInc;
      analogWrite(refAdjPin, vAdj);
      Serial.println("IncreasingVoltage:");
      Serial.print("vAdj= ");
      Serial.println(vAdj * 3.3 / 1024);
    }
  }


  //_____________________________Setup Done__________________________________
  digitalWrite(ledMSPin, LOW);
  digitalWrite(errorLedPin, LOW);
}


//******************************** LOOP *******************************


void loop() {


  //_________________________________Error State_____________________________

  //-----Error State for Changed Gain-----
  //Find out the current gain state:
  if (digitalRead(GSPin) == HIGH) {
    nowGS = 1;
  }
  else if (digitalRead(GSPin) == LOW) {
    nowGS = 2;
  }

  //Find out whether the gain state is still the same:
  if (nowGS != startGS) {
    errorState = true;
  }

  //-----Error State for Removed SD-----
  if (digitalRead(sdDetectPin) == HIGH) {
    errorState = true;
    Serial.println("SD removed in loop");
  }

  //-----Output for Error State-----
  if (errorState == true) {
    msState = false;
    digitalWrite(errorLedPin, HIGH);
    digitalWrite(ledMSPin, LOW);
    Serial.println("ERROR");
  }


  //_________________________Start and Stop Logging__________________________
  
  //-----Start/Stop Button Is Pressed-----
  //Set buttonState true if the button is pressed and button state is false:
  if (digitalRead(measSSPin) == LOW && buttonState == false) {
    delay(30);                                //Wait to de-bounce the switch
    
// Check whether the button is still pressed:
    if (digitalRead(measSSPin) == LOW) {
      buttonState = true;
      
      //-----Start Measurement-----
      //Set the msState high if no measurement is running and there is no 
      //error:
      if (msState == false && errorState == false) {
        msState = true;                          //Enable measuring
        digitalWrite(ledMSPin, HIGH);            //Light measurement LED
      }
      
      //-----Stop Measurement-----
      //Set the msState low if a measurement is running and there is no error
      else if (msState == true && errorState == false) {
        msState = false;                         //Disable measuring 
        digitalWrite(ledMSPin, LOW);             //Switch measurement LED off

        //-----Increase File Name for Next Measurement-----
        //Increase fileNo by one that data is written in one file higher than 
        //the last measurement:
        fileNo = fileNo + 1;
        //Set the measNo back to 1:
        sampleNo = 0;      //Reset sample counter as new file will be used
        measNo = 1;        //That new sub files start from 001

        //Generate the file name:
        if (fileNo < 10) {                                //fileNo [0,9]
          fileName = "00" + String(fileNo) + "00" + String(measNo) + ".txt";
        }
        else if (fileNo < 100 && fileNo >= 10) {          //fileNo [10,99]
          fileName = "0" + String(fileNo) + "00" + String(measNo) + ".txt";
        }
        else if (fileNo < 1000 && fileNo >= 100) {        //fileNo [100,998]
          fileName = String(fileNo) + "00" + String(measNo) + ".txt";
        }

        //Send an error if more than 999 manually started measurements are on
        //the SD card:
        else {                                            //fileNo >999
          Serial.println("More than 999 files on the card");
          errorState = true;
        }
      }
    }
  }

  //-----Start/Stop Button Was Released-----
  //Set buttonState false if the button is not pressed and buttonState true:
  else if (digitalRead(measSSPin) == HIGH && buttonState == true) {
    delay(30);                                //Wait to debounce the switch

    // Check whether the button is still pressed:
    if (digitalRead(measSSPin) == HIGH) {
      buttonState = false;
    }
  }


  //______________________Automatic New File Name___________________________

  if (sampleNo > sampleNoMax - 1) {    //-1 because the counter is increased 
                                       //after writing

    //Increase the measNo by one that the data is stored on a file named
    //higher than the last measurement:
    measNo = measNo + 1;
    sampleNo = 0;        //Reset sample counter because the file will be used

    //Generate the new file name:
    //fileNo [1,9], measNo [1,9]
    if (fileNo < 10 && measNo < 10) {   
      fileName = "00" + String(fileNo) + "00" + String(measNo) + ".txt";
    }
    //fileNo [1,9], measNo [10,99]
    else if (fileNo < 10 && measNo >= 10 && measNo < 100) {               
      fileName = "00" + String(fileNo) + "0" + String(measNo) + ".txt";
    }
    //fileNo [1,9], measNo [100,999]
    else if (fileNo < 10 && measNo >= 100 && measNo < 1000) {                
      fileName = "00" + String(fileNo) + String(measNo) + ".txt";
    }
    //fileNo [10,99], measNo [1,9]
    else if (fileNo >= 10 && fileNo < 100 && measNo < 10) {                      
      fileName = "0" + String(fileNo) + "00" + String(measNo) + ".txt";
    }
    //fileNo [10,99], measNo [10,99]
    else if (fileNo >= 10 && fileNo < 100 && measNo >= 10 && measNo < 100) {             
      fileName = "0" + String(fileNo) + "0" + String(measNo) + ".txt";
    }
    //fileNo [10,99], measNo [100,999]
    else if (fileNo >= 10 && fileNo < 100 && measNo >= 100 && measNo < 1000){           
      fileName = "0" + String(fileNo) + String(measNo) + ".txt";
    }
    //fileNo [100,999], measNo [1,9]
    else if (fileNo >= 100 && fileNo < 1000 && measNo < 10) {                            
      fileName = String(fileNo) + "00" + String(measNo) + ".txt";
    }
    //fileNo [100,999], measNo [10,99]
    else if (fileNo >= 100 && fileNo < 1000 && measNo >= 10 && measNo < 100){           
      fileName = String(fileNo) + "0" + String(measNo) + ".txt";
    }
    //fileNo [100,999], measNo [100,999]
    else if (fileNo >= 100 && fileNo < 1000 && measNo >= 100 && measNo < 
             1000) {         
      fileName = String(fileNo) + String(measNo) + ".txt";
    }

    //Send an error if there are already more than 999 automatically started
    //sub measurements on the SD card:
    else {                          //Error when fileNo > 999 or measNo > 999
      errorState = true;
      Serial.println("Too long measurement (too many sub files)");                       
    }
  }


  //_______________________Multiplex and Analog Read_______________________

  //-----Multiplexer Set 1-----
  //Enable Multiplexer Set 1:
  digitalWrite(E1Pin, LOW);

  //A0 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch01 = analogRead(MSPin);

  //A1 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch02 = analogRead(MSPin);

  //A2 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch03 = analogRead(MSPin);

  //A3 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch04 = analogRead(MSPin);

  //A4 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch05 = analogRead(MSPin);

  //A5 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch06 = analogRead(MSPin);

  //A6 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch07 = analogRead(MSPin);

  //A7 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch08 = analogRead(MSPin);

  //Disable Multiplexer 1
  digitalWrite(E1Pin, HIGH);

  //-----Multiplexer Set 2-----
  //Enable Multiplexer Set 2:
  digitalWrite(E2Pin, LOW);

  //A0 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch09 = analogRead(MSPin);

  //A1 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch10 = analogRead(MSPin);

  //A2 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch11 = analogRead(MSPin);

  //A3 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch12 = analogRead(MSPin);

  //A4 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch13 = analogRead(MSPin);

  //A5 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch14 = analogRead(MSPin);

  //A6 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch15 = analogRead(MSPin);

  //A7 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch16 = analogRead(MSPin);

  //Disable Multiplexer 1
  digitalWrite(E2Pin, HIGH);

  //-----Multiplexer Set 3-----
  //Enable Multiplexer 3 Set:
  digitalWrite(E3Pin, LOW);

  //A0 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch17 = analogRead(MSPin);

  //A1 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch18 = analogRead(MSPin);

  //A2 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch19 = analogRead(MSPin);

  //A3 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, LOW);
  delayMicroseconds(tSet);
  ch20 = analogRead(MSPin);

  //A4 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch21 = analogRead(MSPin);

  //A5 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, LOW);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch22 = analogRead(MSPin);

  //A6 Mux 1 connected:
  digitalWrite(S0Pin, LOW);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch23 = analogRead(MSPin);

  //A7 Mux 1 connected:
  digitalWrite(S0Pin, HIGH);
  digitalWrite(S1Pin, HIGH);
  digitalWrite(S2Pin, HIGH);
  delayMicroseconds(tSet);
  ch24 = analogRead(MSPin);

  //Disable Multiplexer 1
  digitalWrite(E3Pin, HIGH);

  //__________________Write Measured Values on the SD Card___________________

  //Write on the SD card if a measurement is Running:
  if (msState == true) {

    //------Determine the Numbers of Samples on Current File----------
    //Is reset in the block where and when the file number changes
    sampleNo = sampleNo + 1;

    //Open File on the SD card:
    File dataFile = SD.open(fileName, FILE_WRITE);

    //-----Write the Measured Values on the SD Card-----
    //If the file is available, write to it:
    if (dataFile) {
      //Save values on the SD card:
      dataFile.print(millis());
      dataFile.print(",");
      dataFile.print(ch01);
      dataFile.print(",");
      dataFile.print(ch02);
      dataFile.print(",");
      dataFile.print(ch03);
      dataFile.print(",");
      dataFile.print(ch04);
      dataFile.print(",");
      dataFile.print(ch05);
      dataFile.print(",");
      dataFile.print(ch06);
      dataFile.print(",");
      dataFile.print(ch07);
      dataFile.print(",");
      dataFile.print(ch08);
      dataFile.print(",");
      dataFile.print(ch09);
      dataFile.print(",");
      dataFile.print(ch10);
      dataFile.print(",");
      dataFile.print(ch11);
      dataFile.print(",");
      dataFile.print(ch12);
      dataFile.print(",");
      dataFile.print(ch13);
      dataFile.print(",");
      dataFile.print(ch14);
      dataFile.print(",");
      dataFile.print(ch15);
      dataFile.print(",");
      dataFile.print(ch16);
      dataFile.print(",");
      dataFile.print(ch17);
      dataFile.print(",");
      dataFile.print(ch18);
      dataFile.print(",");
      dataFile.print(ch19);
      dataFile.print(",");
      dataFile.print(ch20);
      dataFile.print(",");
      dataFile.print(ch21);
      dataFile.print(",");
      dataFile.print(ch22);
      dataFile.print(",");
      dataFile.print(ch23);
      dataFile.print(",");
      dataFile.print(ch24);
      dataFile.print(",");
      dataFile.println(startGS);
      
      //Close the SD card:
      dataFile.close();
    }

    //If the file cannot be opened, return an error:
    else {
      errorState = true;
      Serial.println("Error opening file!");
    }
  }
}

