%****************************************************************************
%  MATLAB DATA STRESS CALCULATION CODE
%  
%  Version: skiPositionStress
%  Date: 09.10.2016
%  
%  Copyright 2016 by Fabian Tobias Mayer (fabianmayer@web.de)
%  This code may not be used for purposes which are not related to 
%  the master thesis of Fabian Tobias Mayer without his prior
%  written consent. 
%****************************************************************************

function [ sigma_x,sigma_y,sigma_z,tau_xy,tau_xz ] = SkiPositionStress( position,strainRosette,E,nu )
% Summary:
% Calculation of the normal and shear stress in the roller ski based
% coordinate system from the strains in in the rosette systems
%
% Function:
% 1) Calculation of principal stress at the rosette positions with the 
%    measured strain 
% 2) Calculation of the direction of the principal stress:
%    - Calculation of a smallest angle to the a-axis (necessary because of
%      ambiguity of the tangent)
%    - Calculation of the angle between the positive a-axis and the
%      direction of the 1st principal stress (counter clockwise, [rad])
% 3) Calculation of Normal and Shear Stress in the Rosette coordinate system
% 4) Transformation Normal and Shear Stress in Roller Ski coordinate system
%
% Input Variables:
% - position: orientation of the rosette
%     1: Top, pins in positive x-direction
%     2: Top, pins in negative x-direction
%     3: Right, pins in positive z-direction
%     4: Bottom, pins in positive x-direction
%     5: Bottom, pins in negative x-direction
%     6: Left, pins in positive z-direction
%     Else: "Non-existent rosette position": If position for the rosette
%           was not chosen / does not exist
% - epsilo_a, epsilon_b, epsilon_c: strain in the rosettes strain gauges
% - E: Young's modulus in the beam of the roller ski
% - nu: Poisson ratio
%
% Output Variables:
% - sigma_x: Normal stress in positive x direction (skiing direction)
% - sigma_y: Normal stress in positive y direction (horizontal,
%            perpendicular, left in skiing direction
% - sigma_z: Normal stress in positive z direction (vertical, upwards)
% - tau_xy: Shear stress in plane perpendicular to x in positive y
%           direction
% - tau_xz: Shear stress in plane perpendicular to x in positive z
%           direction
 
%% Separate strain vectors for the rosette matrix:
epsilon_a=strainRosette(:,1);
epsilon_b=strainRosette(:,2);
epsilon_c=strainRosette(:,3);
 
%%------------------------Principal Stresses---------------------------------
 
% Number of Measurements (to initialise vector in order to save comp. costs)
n=size(strainRosette,1);
% Initialize vectors for principal stresses:
sigma_1=zeros(n,1);
sigma_2=zeros(n,1);
 
% Calculation of principal stresses [Pa]:
for i=1:n
    sigma_1(i)=E/(2*(1-nu))*(epsilon_a(i)+epsilon_c(i))+E/(sqrt(2)*(1+nu))*sqrt((epsilon_a(i)-epsilon_b(i))^2+(epsilon_c(i)-epsilon_b(i))^2);
    sigma_2(i)=E/(2*(1-nu))*(epsilon_a(i)+epsilon_c(i))-E/(sqrt(2)*(1+nu))*sqrt((epsilon_a(i)-epsilon_b(i))^2+(epsilon_c(i)-epsilon_b(i))^2);
end
 
%%----------------------Direction of Principal Stresses----------------------
 
% Calculation of a smallest angle to the a-axis (necessary because of
% ambiguity of the tangent):
 
numerator=2*epsilon_b-epsilon_a-epsilon_c;
denominator=epsilon_a-epsilon_c;
fraction=numerator./denominator;
 
psi=atan(fraction); % Smallest angle to the a-axis [rad]
 
% Calculation of the angle between the positive a-axis and the
% direction of the 1st principal stress (counter clockwise, [rad]):
 
rho=zeros(n,1); %initialize vector of the angle
 
for i=1:n
    if numerator(i)>=0 && denominator(i)>0
        rho(i)=psi(i)/2;
    elseif numerator(i)>0 && denominator(i)<=0
        rho(i)=(pi-psi(i))/2;
    elseif numerator(i)<=0 && denominator(i)<0
        rho(i)=(pi+psi(i))/2;
    elseif numerator(i)<0 && denominator(i)>=0
        rho(i)=(2*pi-psi(i))/2;
    else
        disp(['Numerator and Denominator Zero (no loads check Set-up)',num2str(i)]);
        rho(i)=0;
    end
end
 
%%--Calculation of Normal and Shear Stress in the Rosette Coordinate System--
 
% Notation from Gross et al. 2009 (Not from HBM!)

% Normal stress in rosette a-axis [pseudo Pa]:
sigma_a=((sigma_1+sigma_2)+(sigma_1-sigma_2).*cos(2*rho))/2;
% Normal stress in rosette c-axis [pseudo Pa]:
sigma_c=((sigma_1+sigma_2)-(sigma_1-sigma_2).*cos(2*rho))/2;
% Shear stress in rosette ac-direction [pseudo Pa]:
tau_ac=-(sigma_1-sigma_2).*sin(2*rho)/2;
 
%%---Transformation Normal and Shear Stress in Roller Ski Coordinate System--
 
% Directions:
%     1: Top, pins in positive x-direction
%     2: Top, pins in negative x-direction
%     3: Right, pins in positive z-direction
%     4: Bottom, pins in positive x-direction
%     5: Bottom, pins in negative x-direction
%     6: Left, pins in positive z-direction
 
% Cases for the possible positions of the rosettes
switch position
    case 1 % Top, pins in positive x-direction
        sigma_x=sigma_c;
        sigma_y=sigma_a;
        sigma_z=zeros(n,1);
        tau_xy=-tau_ac;
        tau_xz=zeros(n,1);
    case 2 % Top, pins in negative x-direction
        sigma_x=sigma_c;
        sigma_y=-sigma_a;
        sigma_z=zeros(n,1);
        tau_xy=-tau_ac;
        tau_xz=zeros(n,1);
    case 3 % Right, pins in positive z-direction
        sigma_x=-sigma_a;
        sigma_y=zeros(n,1);
        sigma_z=-sigma_c;
        tau_xy=zeros(n,1);
        tau_xz=-tau_ac;
    case 4 % Bottom, pins in positive x-direction
        sigma_x=sigma_c;
        sigma_y=-sigma_a;
        sigma_z=zeros(n,1);
        tau_xy=-tau_ac;
        tau_xz=zeros(n,1);
    case 5 % Bottom, pins in negative x-direction
        sigma_x=sigma_c;
        sigma_y=sigma_a;
        sigma_z=zeros(n,1);
        tau_xy=tau_ac;
        tau_xz=zeros(n,1);
    case 6 % Left, pins in positive z-direction
        sigma_x=sigma_a;
        sigma_y=zeros(n,1);
        sigma_z=-sigma_c;
        tau_xy=zeros(n,1);
        tau_xz=-tau_ac;
    
    otherwise % If position for the rosette was not chosen / does not exist
        disp('Non-existent rosette position')
        sigma_x=NaN;
        sigma_y=NaN;
        sigma_z=NaN;
        tau_xy=NaN;
        tau_xz=NaN;
end
 
end
